require 'tamiyo/process'
require 'tamiyo/card'

module Tamiyo
  class ScrapeParser
    include Process::Node

    def pull
      parse_cards_from @chain.xml_spoiler_rows
    end

    def parse_cards_from(xml)
      @collection = {}
      @linked_cards = {}
      xml.each do |tr|
        tds = tr.css 'td'
        if tds.length == 2
          left, right = tds
          right_text = right.content.tr '—', '-'
          case fold left.content
          when 'Name'
            href = right.at_css('a')['href']
            @id = href.match(/multiverseid=(.*)/)[1]
            @name = fold right_text
          when 'Cost:'
            @cost = fold right_text
          when 'Color:'
            @color = fold right_text
          when 'Type:'
            @type = fold right_text
          when 'Pow/Tgh:'
            @pt = fold(right_text).tr '()', ''
            @pt = nil if @pt.empty?
          when 'Loyalty:'
            @loyalty = right_text.strip.tr '()', ''
          when 'Rules Text:'
            @text = right_text.strip.each_line.map(&:strip).join "\n"
          end
        else
          card = parse_card
          @collection[card.name] = card
          @color = @pt = @loyalty = nil
        end
      end
      @collection.values
    end

    def parse_card
      linked_card = @name.match for_linked_card
      @name = beautify @name

      adjusted_card = if linked_card
        part_name = beautify linked_card[:part_name]
        card = @collection[@name]
        if split_card
          adjust_for_split_card card, part_name if card
        else
          adjust_for_top_flip_card card, part_name
        end
      elsif flip_card
        adjust_for_bottom_flip_card
      elsif transform_in_card_text
        adjust_for_transform_card
      end
      adjusted_card || new_card
    end

    def adjust_for_split_card(card, part_name)
      card = if @name.start_with? part_name
        card.with_left_split @cost, @text
      else
        card.with_right_split @cost, @text
      end
      card.with_extra_colors colors
    end

    def adjust_for_top_flip_card(card, bottom_name)
      if card
        card.with_flip_bottom new_card.with_name bottom_name
      else
        remember_top_bottom_link @name, bottom_name
        new_card.with_name bottom_name
      end
    end

    def adjust_for_bottom_flip_card
      bottom_card_name = bottom_card_of @name
      if bottom_card_name
        bottom_card = @collection.delete bottom_card_name
        new_card.with_flip_bottom bottom_card
      end
    end

    def adjust_for_transform_card
      back_name = fetch_back_card_of @name
      new_card.with_transform_hint back_name if back_name
    end

    def fetch_back_card_of(card_name)
      unless remembers_front_back_link card_name
        parts = pull_card_details
        return false if parts == :single
        front_name, back_name = parts
        remember_front_back_link front_name, back_name
        remember_front_back_link back_name, false
      end
      back_card_of card_name
    end

    def pull_card_details
      xml = @chain.xml_card_name_details @id
      return :single if xml.one?
      xml.map do |name_row|
        beautify fold name_row.at_css('.value').content
      end
    end

    def remember_top_bottom_link(first, second)
      @linked_cards.store first, second
    end

    alias :remember_front_back_link :remember_top_bottom_link

    def bottom_card_of(first)
      @linked_cards[first]
    end

    alias :back_card_of :bottom_card_of

    def remembers_front_back_link(first)
      @linked_cards.include? first
    end

    def new_card
      Card.with name: @name,
                cost: @cost,
                type: @type,
                text: @text,
                pt: @pt,
                loyalty: @loyalty,
                colors: colors,
                table_row: table_row,
                played_tapped: played_tapped,
                picture_url: picture_url
    end

    def beautify(name)
      name.gsub(for_linked_card, '')
        .gsub('XX', '')
        .gsub('Æ', 'AE')
        .tr('’', "'")
    end

    def split_card
      @name.include? '//'
    end

    def flip_card
      @text[match_flip_text]
    end

    def transform_in_card_text
      @text[match_transform_text]
    end

    def colors
      @color ||= ''
      {'W' => [' is white.', 'White'],
       'U' => [' is blue.', 'Blue'],
       'B' => [' is black.', 'Black'],
       'R' => [' is red.', 'Red'],
       'G' => [' is green.', 'Green']
      }.each_with_object [] do |(color, (text, identity)), colors|
        if @cost.include?(color) ||
           @text.include?(@name + text) ||
           @color.include?(identity)
          colors << color
        end
      end
    end

    def table_row
      mana_source = @type.end_with?('Artifact') &&
        @text.include?('{T}') && @text.include?('to your mana pool')
      case main_type_based_on @type
      when /Land/, mana_source
        :first
      when /Sorcery|Instant/
        :stack
      when /Creature/
        :third
      else
        :second
      end
    end

    def main_type_based_on(type)
      # The last word before a dash or $
      type.match(/(\w+)\s*(-|$)/)[1]
    end

    def played_tapped
      @text.include?(@name + ' enters the battlefield tapped.')
    end

    def picture_url
      host = 'http://gatherer.wizards.com/'
      path = 'Handlers/Image.ashx'
      query = '?multiverseid=%s&amp;type=card' % @id
      "#{host}#{path}#{query}"
    end

    private

    def fold(text)
      text.strip.gsub /\s+/, ' '
    end

    def match_linked_card
      / \((?<part_name>.*)\)/
    end

    alias :for_linked_card :match_linked_card

    def match_flip_text
      %r{flip (it|#{@name}).|flipped.}
    end

    def match_transform_text
      /transform/i
    end
  end
end