require 'yaml'

module Tamiyo
  module YamlHelper
    def yaml_from(data)
      YAML.load data
    end
  end
  
  module YamlHelper
    def yaml_of(obj)
      YAML.dump obj
    end
  end

  module YamlHelper
    def setup_yaml_emitter_for(file)
      @emit = YAML::Emitter.new file
    end

    def emit_start_of_one_document
      @emit.start_stream Psych::Nodes::Stream::UTF8
      @emit.start_document [1, 1], [], true
    end

    def emit_end_of_one_document
      @emit.end_document true
      @emit.end_stream
    end

    def emit_pair(key, value)
      emit_plain key
      emit_plain value
    end

    def emit_pair_with_literal_value(key, value)
      emit_plain key
      emit_literal value
    end

    def emit_pair_with_optional_sequence_value(key, value)
      if value.kind_of? Array
        emit_pair_with_sequence_value key, value
      else
        emit_pair key, value
      end
    end

    def emit_pair_with_sequence_value(key, value)
      emit_plain key
      emit_flow_sequence do
        value.each { |element| emit_plain element }
      end
    end

    def emit_block_sequence
      @emit.start_sequence nil, nil, true, YAML::Nodes::Sequence::BLOCK
      yield
      @emit.end_sequence
    end

    def emit_flow_sequence
      @emit.start_sequence nil, nil, true, YAML::Nodes::Sequence::FLOW
      yield
      @emit.end_sequence
    end

    def emit_block_mapping
      @emit.start_mapping nil, nil, true, YAML::Nodes::Mapping::BLOCK
      yield
      @emit.end_mapping
    end

    def emit_plain(value)
      @emit.scalar value, nil, nil, true, false, Psych::Nodes::Scalar::PLAIN
    end

    def emit_literal(value)
      @emit.scalar value, nil, nil, false, true, YAML::Nodes::Scalar::LITERAL
    end
  end

  module YamlHelper
    def yaml_event_stream_for(file)
      queue = Queue.new
      Thread.new do
        handler = EventHandler.new queue
        YAML::Parser.new(handler).parse file
      end
      EventStream.new queue
    end

    class EventStream
      attr_reader :type
      attr_reader :value

      def initialize(queue)
        @queue = queue
      end

      def each_within_sequence
        assume_sequence
        until peak_next_is_end
          yield
        end
        assume_end
      end

      def each_within_mapping
        assume_mapping
        until peak_next_is_end
          yield
        end
        assume_end
      end

      def on_types(cases)
        type, _ = peak
        raise StopIteration, "No case for type #{type} available" unless cases.include? type
        self.next if type == :scalar
        cases[type][]
      end

      def on_keys(cases)
        value = assume_scalar
        raise StopIteration, "No case for key #{value} available" unless cases.include? value
        cases[value][]
      end

      def assume_scalar
        self.next
        raise StopIteration, "Assumed scalar but found #{@type}" unless @type == :scalar
        @value
      end

      def assume_sequence
        self.next
        raise StopIteration, "Assumed seq but found #{@type}" unless @type == :seq
      end

      def assume_mapping
        self.next
        raise StopIteration, "Assumed map but found #{@type}" unless @type == :map
      end

      def assume_end
        self.next
        raise StopIteration, "Assumed end but found #{@type}" unless @type == :end
      end

      def peak_next_is_end
        type, _ = peak
        type == :end
      end

      def next
        event = @peak || @queue.pop
        @peak = nil
        @type, @value = event
        event
      end

      def peak
        @peak ||= @queue.pop
      end
    end

    class EventHandler < YAML::Handler
      def initialize(queue)
        @queue = queue
      end

      def scalar(value, anchor, tag, plain, quoted, style)
        @queue << [:scalar, value]
      end

      def start_sequence(anchor, tag, implicit, style)
        @queue << [:seq]
      end

      def start_mapping(anchor, tag, implicit, style)
        @queue << [:map]
      end

      def end_sequence
        @queue << [:end]
      end
      alias :end_mapping :end_sequence
    end
  end
end